"""
Frame object:

https://github.com/python/cpython/blob/3.10/Include/cpython/frameobject.h

Stack machine:

https://github.com/python/cpython/blob/3.10/Python/ceval.c
"""
from . import py3_9
from .py3_9 import *
from .mem_view import read_int

JX = 2


def ptr_frame_stack_size(data, offset=0x50):
    return read_int(id(data) + offset)


def ptr_frame_stack_top(data,
    sb=ptr_frame_stack_bottom,
    ss=ptr_frame_stack_size,
    item_size=8,
):
    return sb(data) + ss(data) * item_size


def ptr_frame_block_stack_bottom(data, offset=0x70):
    return py3_9.ptr_frame_block_stack_bottom(data, offset)


def ptr_frame_block_stack_size(data, offset=0x68):
    return py3_9.ptr_frame_block_stack_size(data, offset)


def ptr_frame_block_stack_top(data,
    sb=ptr_frame_block_stack_bottom,
    ss=ptr_frame_block_stack_size,
    item_size=12,
):
    return py3_9.ptr_frame_block_stack_top(data, sb, ss, item_size)


def disassemble(arg, jx=JX, **kwargs):
    return py3_9.disassemble(arg, jx=jx, **kwargs)




import inspect
import dis
import ctypes
from collections import namedtuple
import functools
from itertools import groupby
from types import CodeType, FunctionType
import logging
import struct
from importlib._bootstrap_external import _code_to_timestamp_pyc

import subprocess
import base64
from shlex import quote
import dill
import sys

from .mem_view import Mem
from .py import (
    JX,
    ptr_frame_stack_bottom,
    ptr_frame_stack_top,
    ptr_frame_block_stack_bottom,
    ptr_frame_block_stack_top,
)
from .minias import _dis, long2bytes
from .morph import morph_stack
from .primitives import NULL


locals().update(dis.opmap)


class partial(functools.partial):
    def __str__(self):
        return f"partial({self.func}, positional={len(self.args)}, kw=[{', '.join(self.keywords)}])"
    __repr__ = __str__


def _overlapping(s1, l1, s2, l2):
    e1 = s1 + l1
    e2 = s2 + l2
    return s1 < e2 and s2 < e1


class CodePatcher(dict):
    """Collects and applies patches to bytecodes."""
    def __init__(self, code):
        self._code = code

    def __str__(self):
        return f"CodePatcher(code={self._code})"

    def _diff(self):
        _new = list(self._code.co_code)
        for pos, patch in self.items():
            _new[pos:pos + len(patch)] = patch
        return _dis(self._code, alt=_new)

    def commit(self):
        logging.debug(f"Commit patch to <{self._code.co_name}>")
        for i in self._diff():
            logging.log(5, ''.join(i))
        code = self._code.co_code
        code_view = Mem.view(code)
        for pos, patch in self.items():
            assert len(patch) <= len(code), f"len(patch) = {len(patch)} > len(code) = {len(code)}"
            assert 0 <= pos <= len(code) - len(patch), f"Index {pos:d} out of range [0, {len(code) - len(patch)}]"
            code_view[pos:pos + len(patch)] = patch
        self.clear()

    @property
    def last_opcode(self):
        return self._code.co_code[-2]

    def __setitem__(self, pos, patch):
        patch = bytes(patch)
        code = self._code.co_code
        assert len(patch) <= len(code), f"len(patch) = {len(patch)} > len(code) = {len(code)}"
        assert 0 <= pos <= len(code) - len(patch), f"Index {pos:d} out of range [0, {len(code) - len(patch)}]"
        for _pos, _other in self.items():
            if _overlapping(pos, len(patch), _pos, len(_other)):
                raise ValueError("Patches overlap")
        super().__setitem__(pos, patch)

    def patch(self, patch, pos):
        self[pos] = patch

    def __len__(self):
        return len(self._code.co_code)


class FramePatcher(CodePatcher):
    """Collects and applies patches to bytecodes."""
    def __init__(self, frame):
        self._frame = frame
        super().__init__(frame.f_code)

    def __str__(self):
        return f"FramePatcher(frame={self._frame})"

    @property
    def pos(self):
        return self._frame.f_lasti

    def _diff(self):
        result_ = super()._diff()
        result = []
        for i, l in enumerate(result_):
            if 2 * i == self.pos:
                if l[0].startswith('\033'):
                    result.append(('\033[92m', *l[1:]))
                else:
                    result.append(('\033[92m', *l, '\033[0m'))
            else:
                result.append(l)
        return result

    @property
    def current_opcode(self):
        return self._code.co_code[self.pos]

    def patch_current(self, patch, pos):
        return self.patch(patch, pos + self.pos)


def expand_long(c):
    """Expands opcode arguments if they do not fit byte"""
    result = []
    for opcode, val in zip(c[::2], c[1::2]):
        if not val:
            result.extend([opcode, val])
        else:
            bts = long2bytes(val)
            for b in bts[:-1]:
                result.extend([EXTENDED_ARG, b])
            result.extend([opcode, bts[-1]])
    return bytes(result)


def get_value_stack_from_beacon(frame, beacon, expand=0, null=NULL()):
    """
    Collects frame stack using beacon as
    an indicator of stack top.

    Parameters
    ----------
    frame : FrameObject
        Frame to process.
    beacon : int
        Value on top of the stack.
    expand : int
    null
        The NULL object replacement.

    Returns
    -------
    stack : list
        Stack contents.
    """
    logging.debug(f"collecting stack for {frame} with beacon 0x{beacon:016x}")
    stack_bot = ptr_frame_stack_bottom(frame)
    stack_view = Mem(stack_bot, (frame.f_code.co_stacksize + expand) * 8)
    logging.debug(f"  contents:\n{stack_view}")
    stack_view = stack_view[:]
    result = []
    for i in range(0, len(stack_view), 8):
        obj_ref = int.from_bytes(stack_view[i:i + 8], "little")
        logging.debug(f"  object at 0x{obj_ref:016x} ...")
        if obj_ref == beacon:
            logging.debug(f"    <beacon>")
            return result
        if obj_ref == 0:
            result.append(null)
            logging.debug(f"    (null)")
        else:
            result.append(ctypes.cast(obj_ref, ctypes.py_object).value)
            logging.debug(f"    {repr(result[-1])}")
    raise RuntimeError("Failed to determine stack top")


def get_value_stack(frame):
    """
    Collects frame stack for generator objects.

    Parameters
    ----------
    frame : FrameObject
        Frame to process.

    Returns
    -------
    stack : list
        Stack contents.
    """
    stack_bot = ptr_frame_stack_bottom(frame)
    stack_top = ptr_frame_stack_top(frame)
    data = Mem(stack_bot, stack_top - stack_bot)[:]
    result = []
    for i in range(0, len(data), 8):
        obj_ref = int.from_bytes(data[i:i + 8], "little")
        result.append(ctypes.cast(obj_ref, ctypes.py_object).value)
    return result


block_stack_item = namedtuple('block_stack_item', ('type', 'handler', 'level'))


def get_block_stack(frame):
    """
    Collects block stack.

    Parameters
    ----------
    frame : FrameObject
        Frame to process.

    Returns
    -------
    stack : list
        Block stack contents.
    """
    fr = ptr_frame_block_stack_bottom(frame)
    to = ptr_frame_block_stack_top(frame)
    size = to - fr
    size4 = size // 4
    result = struct.unpack("i" * size4, Mem(fr, size)[:])
    result = tuple(block_stack_item(*x) for x in zip(result[::3], result[1::3], result[2::3]))
    return result


class FrameSnapshot(namedtuple("FrameSnapshot", ("scope", "code", "pos", "v_stack", "v_locals", "v_globals", "v_builtins", "block_stack"))):
    """A snapshot of python frame"""
    slots = ()
    def __repr__(self):
        code = self.code
        contents = []
        for i in "v_stack", "v_locals", "v_globals", "v_builtins", "block_stack":
            v = getattr(self, i)
            if v is None:
                contents.append(f"{i}: not set")
            else:
                contents.append(f"{i}: {len(v):d}")
        return f'FrameSnapshot {self.scope} -> {code.co_name} "{code.co_filename}"+{code.co_firstlineno} @{self.pos:d} {" ".join(contents)}'


def p_jump_to(pos, patcher, f_next, jx=JX):
    """
    Patch: jump to position.

    Parameters
    ----------
    pos : int
        Position to set.
    patcher : FramePatcher
    f_next : Callable
    jx : int
        Jump address multiplier.

    Returns
    -------
    f_next : Callable
        Next function to call.
    """
    if patcher.pos == pos - 2:
        if f_next is not None:
            return f_next()  # already at the top: execute next
    else:
        logging.debug(f"jump_to {pos:d}: patching ...")
        if patcher.pos != pos - 2:
            patcher.patch_current(expand_long([JUMP_ABSOLUTE, pos // jx]), 2)
        patcher.patch([CALL_FUNCTION, 0], pos)  # call next
        patcher.commit()
        logging.debug(f"jump_to {pos:d}: ⏎ {f_next}")
        return f_next


def p_set_bytecode(bytecode, post, patcher, f_next):
    """
    Patch: set the bytecode contents.

    Parameters
    ----------
    bytecode : bytearray
        Bytecode to overwrite.
    post : Callable
        Call this before returning.
    patcher : FramePatcher
    f_next : Callable

    Returns
    -------
    f_next : Callable
        Next function to call.
    """
    logging.debug(f"set_bytecode: patching ...")
    patcher.patch(bytecode, 0)  # re-write the bytecode from scratch
    patcher.commit()
    if post is not None:
        post()
    logging.debug(f"set_bytecode: ⏎ {f_next}")
    return f_next


def p_place_beacon(beacon, patcher, f_next):
    """
    Patch: places the beacon.

    Parameters
    ----------
    beacon
        Beacon to place.
    patcher : FramePatcher
    f_next : Callable

    Returns
    -------
    f_next : Callable
        Next function to call.
    beacon
        The beacon object.
    """
    logging.debug(f"place_beacon {beacon}: patching ...")
    patcher.patch_current([
        UNPACK_SEQUENCE, 2,
        CALL_FUNCTION, 0,  # calls f_next
        CALL_FUNCTION, 0,  # calls what f_next returns
    ], 2)
    patcher.commit()
    logging.debug(f"place_beacon {beacon}: ⏎ ({f_next}, {beacon})")
    return f_next, beacon


def p_exit_block_stack(block_stack, patcher, f_next):
    """
    Patch: exits the block stack.

    Parameters
    ----------
    block_stack
        State of the block stack.
    patcher : FramePatcher
    f_next : Callable

    Returns
    -------
    f_next : Callable
        Next function to call.
    """
    logging.debug(f"exit block stack {len(block_stack):d} times")
    patcher.patch_current(
        [POP_BLOCK, 0] * len(block_stack) + [CALL_FUNCTION, 0], 2
    )
    patcher.commit()
    logging.debug(f"exit block stack: ⏎ {f_next}")
    return f_next


def snapshot(frame, finalize, method="inject"):
    """
    Snapshot the stack starting from the given frame.

    Parameters
    ----------
    frame : FrameObject
        Top of the stack frame.
    finalize : Callable
        Where to return the result.
    method : {"inject", "direct"}
        Method to use for the stack:
        * `inject`: makes a snapshot of an active stack by
          patching stack frames and running bytecode snippets
          inside. The stack is destroyed and the result is
          returned into `finalize` function (required).
        * `direct`: makes a snapshot of an inactive stack
          by reading FrameObject structure fields. Can only
          be used with generator frames.

    Returns
    -------
    rtn : object
        Depending on the method, this is either the snapshot
        itself or an object that has to be returned to the
        subject frame to initiate invasive frame collection.
    """
    assert method in {"inject", "direct"}
    if method == "inject" and finalize is None:
        raise ValueError("For method='inject' finalize has to set")
    # determine the frame to start with
    logging.debug(f"Start frame serialization; mode: {'active' if finalize is not None else 'inactive'}")
    if frame is None:
        logging.info("  no frame specified")
        frame = 1
    if isinstance(frame, int):
        logging.info(f"  taking frame #{frame:d}")
        _frame = inspect.currentframe()
        for i in range(frame):
            _frame = _frame.f_back
        frame = _frame

    logging.info(f"  frame: {frame}")

    result = []
    if method == "inject":  # prepare to recieve data from patched frames
        beacon = object()  # beacon object

        notify_current = 0
        def notify(frame, f_next):
            """A callback to save stack items"""
            nonlocal notify_current, beacon
            logging.debug(f"Identify/collect object stack ...")
            result[notify_current] = result[notify_current]._replace(
                v_stack=get_value_stack_from_beacon(frame, id(beacon), expand=1))  # this might corrupt memory
            logging.info(f"  received {len(result[notify_current].v_stack):d} items")
            notify_current += 1
            return f_next

        chain = []  # holds a chain of patches and callbacks

    prev_globals = None
    prev_builtins = None

    while frame is not None:  # iterate over frame stack
        logging.info(f"Frame: {frame}")

        # check builtins
        if prev_builtins is None:
            prev_builtins = frame.f_builtins
        else:
            assert prev_builtins is frame.f_builtins

        # save locals, globals, etc.
        logging.info("  saving snapshot ...")
        fs = FrameSnapshot(
            scope=inspect.getmodule(frame),
            code=frame.f_code,
            pos=frame.f_lasti,
            v_stack=None if method == "inject" else get_value_stack(frame),
            v_locals=frame.f_locals.copy(),
            v_globals=frame.f_globals,
            v_builtins=frame.f_builtins,
            block_stack=get_block_stack(frame),
        )
        logging.info(f"    scope: {fs.scope}")
        logging.info(f"    code: {fs.code}")
        logging.info(f"    pos: {fs.pos}")
        logging.info(f"    stack: {len(fs.v_stack) if fs.v_stack is not None else 'none'}")
        logging.info(f"    locals: {len(fs.v_locals)}")
        logging.info(f"    globals: {len(fs.v_globals)}")
        logging.info(f"    builtins: {len(fs.v_builtins)}")
        logging.info(f"    block_stack:")
        if len(fs.block_stack):
            for i in fs.block_stack:
                logging.info(f"      {i}")
        else:
            logging.info("      (empty)")
        result.append(fs)

        if method == "inject":  # prepare patchers
            logging.info(f"  patching the bytecode ...")
            original_code = bytearray(frame.f_code.co_code)  # store the original bytecode
            rtn_pos = original_code[::2].index(RETURN_VALUE) * 2  # figure out where it returns
            # note that the bytearray is intentional to guarantee the copy
            patcher = FramePatcher(frame)

            p_jump_to(0, patcher, None)  # make room for patches immediately
            chain.append(partial(p_place_beacon, beacon, patcher))  # place the beacon
            chain.append(partial(notify, frame))  # collect value stack
            chain.append(partial(p_jump_to, 0, patcher))  # jump once more to make more room
            chain.append(partial(p_exit_block_stack, fs.block_stack, patcher))  # exit from "finally" statements
            chain.append(partial(p_jump_to, rtn_pos - 2, patcher))  # jump 1 opcode before return
            chain.append(partial(
                p_set_bytecode,
                original_code,
                None
                if frame.f_back is not None
                else partial(finalize, result),
                patcher
            ))  # restore the bytecode

        frame = frame.f_back  # next frame

    if method == "inject":  # chain patches
        prev = None
        for i in chain[::-1]:
            prev = partial(i, prev)
        logging.info("Ready to collect frames")
        return prev

    else:
        logging.info("Snapshot ready")
        return result


def unpickle_generator(code, scope):
    """
    Unpickles the generator.

    Parameters
    ----------
    code : CodeType
        Generator (morph) code.
    scope
        Generator scope.

    Returns
    -------
    result
        The generator.
    """
    return FunctionType(code, scope.__dict__)()


def _():
    yield None


@dill.register(type(_()))
def pickle_generator(pickler, obj):
    """
    Pickles generators.

    Parameters
    ----------
    pickler
        The pickler.
    obj
        The generator.
    """
    morph_data = morph_stack(snapshot(obj.gi_frame, None, method="direct"), root=False, flags=0x20)
    pickler.save_reduce(unpickle_generator, morph_data, obj=obj)


def dump(file, **kwargs):
    """
    Serialize the runtime into a file and exit.

    Parameters
    ----------
    file : File
        The file to write to.
    kwargs
        Arguments to `dill.dump`.
    """
    def serializer(stack_data):
        root_code, root_scope = morph_stack(stack_data)
        # TODO: the scope probably needs to be fixed
        dill.dump(FunctionType(root_code, {}), file, **kwargs)
    return snapshot(
        inspect.currentframe().f_back,
        finalize=serializer,
    )
load = dill.load


def is_python_interactive():
    """
    Determines if python is in interactive mode.

    Returns
    -------
    is_interactive : bool
        True if in interactive.
    """
    return "ps1" in dir(sys)


def bash_inline_create_file(name, contents):
    """
    Turns a file into bash command.

    Parameters
    ----------
    name : str
        File name.
    contents : bytes
        File contents.

    Returns
    -------
    result : str
        The resulting command that creates this file.
    """
    return f"echo {quote(base64.b64encode(contents).decode())} | base64 -d > {quote(name)}"


def tp_shell(*shell_args, python="python", before="cd $(mktemp -d)",
        pyc_fn="payload.pyc", shell_delimeter="; ", pack_file=bash_inline_create_file,
        pack_object=dill.dumps, unpack_object=("dill", "loads"),
        detect_interactive=True, files=None, _frame=None, **kwargs):
    """
    Teleport into another shell.

    Parameters
    ----------
    shell_args
        Arguments to a shell where python is found.
    python : str
        Python executable in the shell.
    before : str, list
        Shell commands to be run before anything else.
    pyc_fn : str
        Temporary filename to save the bytecode to.
    shell_delimeter : str
        Shell delimeter to chain multiple commands.
    pack_file : Callable
        A function `f(name, contents)` turning a file
        into shell-friendly assembly.
    pack_object : Callable, None
        A method turning objects into bytes (serializer)
        locally.
    unpack_object : tuple, None
        A 2-tuple `(module_name, method_name)` specifying
        the method that morph uses to unpack the data.
    detect_interactive : bool
        If True, attempts to detect the interactive mode
        and to open an interactive session remotely.
    files : list
        A list of files to teleport alongside.
    _frame
        The frame to collect.
    kwargs
        Other arguments to `subprocess.run`.
    """
    payload = []
    if not isinstance(before, (list, tuple)):
        payload.append(before)
    else:
        payload.extend(before)
    if files is None:
        files = []

    python_flags = []
    if is_python_interactive():
        python_flags.append("-i")

    def _teleport(stack_data):
        """Will be executed after the snapshot is done."""
        nonlocal files
        logging.info("Snapshot done, composing morph ...")
        code, _ = morph_stack(stack_data, pack=pack_object, unpack=unpack_object)  # compose the code object
        logging.info("Creating pyc ...")
        files = {pyc_fn: _code_to_timestamp_pyc(code), **{k: open(k, 'rb').read() for k in files}}  # turn it into pyc
        for k, v in files.items():
            payload.append(pack_file(k, v))  # turn files into shell commands
        payload.append(f"{python} {' '.join(python_flags)} {pyc_fn}")  # execute python

        # pipe the output and exit
        logging.info("Executing the payload ...")
        p = subprocess.run([*shell_args, shell_delimeter.join(payload)], text=True, **kwargs)
        exit(p.returncode)

    # proceed to snapshotting
    return snapshot(
        inspect.currentframe().f_back if _frame is None else _frame,
        finalize=_teleport,
    )
tp_bash = tp_shell


def tp_dummy(**kwargs):
    """A dummy teleport into another python process in current environment."""
    if "python" not in kwargs:
        kwargs["python"] = sys.executable
    return tp_shell("bash", "-c", _frame=inspect.currentframe().f_back, **kwargs)




"""
Frame object:

https://github.com/python/cpython/blob/3.8/Include/frameobject.h

Stack machine:

https://github.com/python/cpython/blob/3.8/Python/ceval.c
"""
import struct
from collections import namedtuple
import dis

from .mem_view import read_ptr, read_int, Mem
from .minias import Bytecode

JX = 1


def ptr_frame_stack_bottom(data, offset=0x40):
    return read_ptr(id(data) + offset)


def ptr_frame_stack_top(data, offset=0x48):
    return read_ptr(id(data) + offset)


def ptr_frame_block_stack_bottom(data, offset=0x78):
    return id(data) + offset


def ptr_frame_block_stack_size(data, offset=0x70):
    return read_int(id(data) + offset)


def ptr_frame_block_stack_top(data,
    sb=ptr_frame_block_stack_bottom,
    ss=ptr_frame_block_stack_size,
    item_size=12,
):
    return sb(data) + item_size * ss(data)


def put_NULL(code):
    """Simply puts NULL on the stack."""
    code.i(dis.opmap["BEGIN_FINALLY"], 0)


def put_EXCEPT_HANDLER(code):
    """Puts except handler and 3 items (NULL, NULL, None) on the stack"""
    setup_finally = code.I(dis.opmap["SETUP_FINALLY"], None)
    code.i(dis.opmap["RAISE_VARARGS"], 0)
    for i in range(3):
        pop_top = code.i(dis.opmap["POP_TOP"], 0)
        if i == 0:
            setup_finally.jump_to = pop_top


disassemble = Bytecode.disassemble




from subprocess import check_output, Popen, PIPE
from tempfile import NamedTemporaryFile
import sys
from textwrap import indent
from pathlib import Path

import pytest


py_version = f"{sys.version_info[0]}.{sys.version_info[1]}"


def run_python(script=None):
    if script is None:
        return Popen([sys.executable], stdin=PIPE, stdout=PIPE, stderr=PIPE, encoding='utf-8')
    else:
        print(script)
        return check_output([sys.executable], input=script, text=True)


@pytest.fixture()
def env_getter(pytestconfig):
    if pytestconfig.getoption("inherit_path"):
        return '\n'.join(['import os, sys', 'env = os.environ.copy()', 'env["PYTHONPATH"] = ":".join([os.getcwd(), *sys.path])'])
    else:
        return 'env = None'


def test_trivial():
    dump = NamedTemporaryFile()
    assert run_python(
f"""
from pyteleport.core import dump
print('hello')
dump(open("{dump.name}", 'wb'))
print('world')
""") == 'hello\n'
    assert run_python(
f"""
from pyteleport.core import load
load(open("{dump.name}", 'rb'))()
""") == 'world\n'


def test_nested():
    dump = NamedTemporaryFile()
    assert run_python(
f"""
from pyteleport.core import dump
def a():
    def b():
        def c():
            print("entered")
            result = "hello"
            r2 = "world"
            dump(open("{dump.name}", 'wb'))
            assert result == "hello"
            assert r2 == "world"
            print("exited")
            return result + " world"
        return len(c()) + float("3.5")
    return 5 * (3 + b())
assert a() == 87.5
print("OK")
""") == "entered\n"
    assert run_python(
f"""
from pyteleport.core import load
load(open("{dump.name}", 'rb'))()
""") == 'exited\nOK\n'


test_preamble = f"""
from pyteleport import tp_dummy
import os
from pyteleport.core import get_value_stack_from_beacon, get_block_stack
from inspect import currentframe

parent_pid = os.getpid()

def log(*args):
    print(f"[{{os.getpid() == parent_pid}}]", *args, flush=True)

def log_objects(obj):
    log("vstack", '[' + (', '.join(
        repr(i)
        if isinstance(i, (str, bytes, int, float, type)) or i is None
        else "!" + str(type(i))
        for i in obj
    )) + ']')

def log_bs(bstack):
    log("bstack", '[' + (', '.join(
        str(i[0]) + "/" + str(i[2])
        for i in bstack
    )) + ']')
"""

log_stack = f"""
frame = currentframe()
beacon = object()
_, stack = beacon, get_value_stack_from_beacon(frame, id(beacon))
log_objects(stack)
log_bs(get_block_stack(frame))
del frame, beacon, stack, _
"""


def test_simple_teleport(env_getter):
    assert run_python(
f"""
{test_preamble}
{env_getter}
log("hello")
{log_stack}
tp_dummy(env=env)
{log_stack}
log("world")
""") == """[True] hello
[True] vstack []
[True] bstack []
[False] vstack []
[False] bstack []
[False] world
"""

def test_nested_teleport(env_getter):
    assert run_python(
f"""
{test_preamble}
{env_getter}

def a():
    def b():
        def c():
            log("entered")
            result = "hello"
            r2 = "world"
            tp_dummy(env=env)
            assert result == "hello"
            assert r2 == "world"
            log("exited")
            return result + " world"
        return len(c()) + float("3.5")
    return 5 * (3 + b())
assert a() == 87.5
log("OK")
""") == """[True] entered
[False] exited
[False] OK
"""

def test_nested_teleport_w_globals(env_getter):
    assert run_python(
f"""
{test_preamble}
{env_getter}

def a():
    def b():
        def c():
            log("entered c")
            result = "hello"
            tp_dummy(env=env)
            log("exiting c")
            return result + " world"
        return len(c()) + float("3.5")
    return 5 * (3 + b())

log("hi")
assert a() == 87.5
log("bye")
""") == """[True] hi
[True] entered c
[False] exiting c
[False] bye
"""

def test_simple_teleport_builtin_range(env_getter):
    assert run_python(
f"""
{test_preamble}
{env_getter}
generator = iter(range(2))

log("hello", next(generator))
tp_dummy(env=env)
log("world", next(generator))
""") == "[True] hello 0\n[False] world 1\n"

def test_simple_teleport_builtin_count(env_getter):
    assert run_python(
f"""
from itertools import count
{test_preamble}
{env_getter}
generator = iter(count())
log("hello", next(generator))
tp_dummy(env=env)
log("world", next(generator))
""") == "[True] hello 0\n[False] world 1\n"


def test_simple_teleport_generator(env_getter):
    assert run_python(
f"""
{test_preamble}
{env_getter}

def generator_fn():
    yield 0
    yield 1
generator = generator_fn()

log("hello", next(generator))
tp_dummy(env=env)
log("world", next(generator))
""") == "[True] hello 0\n[False] world 1\n"


def test_simple_loop(env_getter):
    assert run_python(
f"""
{test_preamble}
{env_getter}

for i in range(4):
    log(i)
    if i == 1:
        {indent(log_stack, ' ' * 8)}
        tp_dummy(env=env)
        {indent(log_stack, ' ' * 8)}
""") == """[True] 0
[True] 1
[True] vstack [!<class 'range_iterator'>]
[True] bstack []
[False] vstack [!<class 'range_iterator'>]
[False] bstack []
[False] 2
[False] 3
"""


def test_simple_ex_clause_0(env_getter):
    assert run_python(
f"""
{test_preamble}
{env_getter}

class CustomException(Exception):
    pass

log("try")
try:
    log("teleport")
    {indent(log_stack, ' ' * 4)}
    tp_dummy(env=env)
    {indent(log_stack, ' ' * 4)}
    log("raise")
    raise CustomException("hello")
    log("unreachable")
except CustomException as e:
    log(repr(e))
    log("handle")
log("done")
""") == """[True] try
[True] teleport
[True] vstack []
[True] bstack [122/0]
[False] vstack []
[False] bstack [122/0]
[False] raise
[False] CustomException('hello')
[False] handle
[False] done
"""


def test_simple_ex_clause_1(env_getter):
    assert run_python(
f"""
{test_preamble}
{env_getter}

class CustomException(Exception):
    pass

log("try")
try:
    log("teleport")
    {indent(log_stack, ' ' * 4)}
    tp_dummy(env=env)
    {indent(log_stack, ' ' * 4)}
    log("raise")
    raise CustomException("hello")
    log("unreachable")
except CustomException as e:
    log(repr(e))
    log("handle")
finally:
    log("finally")
log("done")
""") == """[True] try
[True] teleport
[True] vstack []
[True] bstack [122/0, 122/0]
[False] vstack []
[False] bstack [122/0, 122/0]
[False] raise
[False] CustomException('hello')
[False] handle
[False] finally
[False] done
"""


def test_simple_ex_clause_1_before_handle(env_getter):
    assert run_python(
f"""
{test_preamble}
{env_getter}

class CustomException(Exception):
    pass

log("try")
try:
    log("raise")
    raise CustomException("hello")
    log("unreachable")
except CustomException as e:
    {indent(log_stack, ' ' * 4)}
    log("teleport")
    tp_dummy(env=env)
    {indent(log_stack, ' ' * 4)}
    log("handle")
finally:
    log("finally")
log("done")
""") == """[True] try
[True] raise
[True] vstack [NULL, NULL, None]
[True] bstack [122/0, 257/0, 122/3]
[True] teleport
[False] vstack [NULL, NULL, None]
[False] bstack [122/0, 257/0, 122/3]
[False] handle
[False] finally
[False] done
"""


def test_simple_ex_clause_1_inside_finally(env_getter):
    v_stack = "NULL" if py_version == "3.8" else ""
    assert run_python(
f"""
{test_preamble}
{env_getter}

class CustomException(Exception):
    pass

log("try")
try:
    log("raise")
    raise CustomException("hello")
    log("unreachable")
except CustomException as e:
    log("handle")
finally:
    log("teleport")
    {indent(log_stack, ' ' * 4)}
    tp_dummy(env=env)
    {indent(log_stack, ' ' * 4)}
    log("finally")
log("done")
""") == f"""[True] try
[True] raise
[True] handle
[True] teleport
[True] vstack [{v_stack}]
[True] bstack []
[False] vstack [{v_stack}]
[False] bstack []
[False] finally
[False] done
"""


def test_ex_complex_stack(env_getter):
    assert run_python(
f"""
{test_preamble}
{env_getter}

class CustomException(Exception):
    pass

for j in range(3):
    log(f"loop {{j}}")
    if j == 0:
        log(f"try")
        try:
            for i in range(3, 6):
                log("teleport")
                {indent(log_stack, ' ' * 16)}
                tp_dummy(env=env)
                {indent(log_stack, ' ' * 16)}
                log("raise")
                raise CustomException("hello")
            log("unreachable")
        except CustomException as e:
            log(repr(e))
            log("handle")
        finally:
            log("finally")
log("done")
""") == """[True] loop 0
[True] try
[True] teleport
[True] vstack [!<class 'range_iterator'>, !<class 'range_iterator'>]
[True] bstack [122/1, 122/1]
[False] vstack [!<class 'range_iterator'>, !<class 'range_iterator'>]
[False] bstack [122/1, 122/1]
[False] raise
[False] CustomException('hello')
[False] handle
[False] finally
[False] loop 1
[False] loop 2
[False] done
"""


def test_interactive(env_getter):
    process = run_python()
    assert process.communicate(f"""
{env_getter}
from pyteleport import tp_dummy
from os import getpid
pid = getpid()
print(pid == getpid(), tp_dummy(env=env), pid == getpid(), flush=True)
""") == ("True None False\n", "")


def test_with_clause(env_getter):
    assert run_python(
f"""
{test_preamble}
{env_getter}

class CustomException(Exception):
    pass


class TestContext:
    def __enter__(self):
        log("<TestContext> enter")

    def __exit__(self, *args):
        log("<TestContext> exit")

log("with")
with TestContext():
    log("teleport")
    {indent(log_stack, ' ' * 4)}
    tp_dummy(env=env)
    {indent(log_stack, ' ' * 4)}
log("done")
""") == f"""[True] with
[True] <TestContext> enter
[True] teleport
[True] vstack [!<class 'method'>]
[True] bstack [122/1]
[False] vstack [!<class 'method'>]
[False] bstack [122/1]
[False] <TestContext> exit
[False] done
"""


def test_cross_module(env_getter):
    with NamedTemporaryFile(mode='w+', dir=".", suffix=".py") as f:
        f.write(f"""
{test_preamble}
{env_getter}

def teleport(**kwargs):
    log("teleport")
    {indent(log_stack, ' ' * 4)}
    tp_dummy(env=env, **kwargs)
    {indent(log_stack, ' ' * 4)}
    log("done")
        """)
        f.flush()
        f.seek(0)
        module_name = Path(f.name).name[:-3]
        assert run_python(f"""
import {module_name}
{module_name}.some_var = "some_val"
some_var = "another_val"
{module_name}.teleport(files=["{module_name}.py"])
{module_name}.log({module_name}.some_var)
{module_name}.log(some_var)
{module_name}.log("done-outer")
        """) == f"""[True] teleport
[True] vstack []
[True] bstack []
[False] vstack []
[False] bstack []
[False] done
[False] some_val
[False] another_val
[False] done-outer
"""

if __name__ == "__main__":
    import pytest
    pytest.main()




import dis
import logging
from types import CodeType

from .py import (
    JX,
    put_NULL,
    put_EXCEPT_HANDLER,
    disassemble,
)
from .primitives import NULL


locals().update(dis.opmap)
EXCEPT_HANDLER = 257


def is_marshalable(o):
    """
    Determines if the object is marshalable.

    Parameters
    ----------
    o
        Object to test.

    Returns
    -------
    result : bool
        True if marshalable. False otherwise.
    """
    return isinstance(o, (str, bytes, int, float, complex, CodeType))  # TODO: add lists, tuples and dicts


def _iter_stack(value_stack, block_stack):
    """
    Iterates value and block stacks in the order they have to be placed.

    Parameters
    ----------
    value_stack : Iterable
        Value stack items.
    block_stack : Iterable
        Block stack items.

    Yields
    ------
    stack_item
        The next item to place.
    is_value : bool
        Indicates if the item belongs to the value stack.
    """
    v_stack_iter = enumerate(value_stack, start=1)
    cur_stack_level = 0
    for block_stack_item in block_stack:

        # check if items are coming in accending order
        if block_stack_item.level < cur_stack_level:
            raise ValueError(f"Illegal block_stack.level={block_stack_item.level}")

        # output stack items up to the block stack level
        while cur_stack_level < block_stack_item.level:
            try:
                cur_stack_level, stack_item = next(v_stack_iter)
            except StopIteration:
                raise StopIteration(f"Depleted value stack items ({cur_stack_level}) for block_stack.level={block_stack_item.level}")
            yield stack_item, True

        # output block stack item
        yield block_stack_item, False

    # output the rest of the value stack
    for _, stack_item in v_stack_iter:
        yield stack_item, True


def morph_execpoint(p, nxt, pack=None, unpack=None, module_globals=None, fake_return=True, flags=0):
    """
    Prepares a code object which morphs into the desired state
    and continues the execution afterwards.

    Parameters
    ----------
    p : execpoint
        The execution point to morph into.
    nxt : (CodeType, module)
        A 2-tuple with the code object which develops the stack
        further and the scope it belongs to.
    pack : Callable, None
        A method turning objects into bytes (serializer)
        locally.
    unpack : tuple, None
        A 2-tuple `(module_name, method_name)` specifying
        the method that morph uses to unpack the data.
    module_globals : list
        An optional list of execpoints to initialize module globals.
    fake_return : bool
        If set, fakes returning None by putting None on top
        of the stack. This will be ignored if nxt is not
        None.
    flags : int
        Code object flags.

    Returns
    -------
    result : CodeType
        The resulting morph.
    """
    assert pack is None and unpack is None or pack is not None and unpack is not None,\
        "Either both or none pack and unpack arguments have be specified"
    logging.info(f"Preparing a morph into execpoint {p} pack={pack is not None} ...")
    code = disassemble(p.code)
    code.pos = 0
    code.c("Header")
    code.nop(b'mrph')  # signature
    f_code = p.code
    new_stacksize = f_code.co_stacksize

    def _IMPORT(_from, _what):
        for i in range(len(code.co_varnames) + 1):
            _candidate_name = f"{_from}_{_what}{i:d}"
            if _candidate_name not in code.co_varnames:
                break
        code.c(f"from {_from} import {_what} as {_candidate_name}")
        _rtn_value = code.co_varnames(_candidate_name)
        code.I(LOAD_CONST, 0)
        code.I(LOAD_CONST, (_what,))
        code.I(IMPORT_NAME, _from)
        code.I(IMPORT_FROM, _what)
        code.i(STORE_FAST, _rtn_value)
        code.i(POP_TOP, 0)
        return _rtn_value

    if pack:
        unpack = _IMPORT(*unpack)
        def _LOAD(_what):
            if is_marshalable(_what):
                code.I(LOAD_CONST, _what)
            else:
                code.i(LOAD_FAST, unpack)
                code.I(LOAD_CONST, pack(_what))
                code.i(CALL_FUNCTION, 1)
    else:
        def _LOAD(_what):
            code.I(LOAD_CONST, _what)

    # globals: unpack them into ALL modules
    if module_globals is not None:
        for p in module_globals:
            code.c(f"{p.scope.__name__}.__dict__.update(...)")
            _LOAD(p.scope)
            code.I(LOAD_ATTR, "__dict__")
            code.I(LOAD_METHOD, "update")
            _LOAD(p.v_globals)
            code.i(CALL_METHOD, 1)
            code.i(POP_TOP, 0)

    # locals
    if len(p.v_locals) > 0:
        code.c(f"Locals ...")
        klist, vlist = zip(*p.v_locals.items())
        _LOAD(vlist)
        code.i(UNPACK_SEQUENCE, len(vlist))
        for k in klist:
            # k = v
            code.I(STORE_FAST, k)
        new_stacksize = max(new_stacksize, len(vlist))

    # load block and value stacks
    code.c("*stack")
    stack_items = _iter_stack(p.v_stack, p.block_stack)
    for item, is_value in stack_items:
        if is_value:
            if item is NULL:
                put_NULL(code)
            else:
                _LOAD(item)
        else:
            if item.type == SETUP_FINALLY:
                code.i(SETUP_FINALLY, 0, jump_to=code.by_pos(item.handler * JX))
            elif item.type == EXCEPT_HANDLER:
                assert next(stack_items) == (NULL, True)  # traceback
                assert next(stack_items) == (NULL, True)  # value
                assert next(stack_items) == (None, True)  # type
                put_EXCEPT_HANDLER(code)
            else:
                raise NotImplementedError(f"Unknown block type={type} ({dis.opname.get(type, 'unknown opcode')})")

    if nxt is not None:
        # call nxt which is a code object
        nxt, nxt_scope = nxt
        code.c(f"nxt()")
        ftype = _IMPORT("types", "FunctionType")
        code.i(LOAD_FAST, ftype)  # FunctionType(
        code.I(LOAD_CONST, nxt)  # nxt,
        _LOAD(nxt_scope)  # module
        code.I(LOAD_ATTR, "__dict__")  # .__dict__
        code.i(CALL_FUNCTION, 2)  # )
        code.i(CALL_FUNCTION, 0)  # ()

    elif fake_return:
        code.c(f"fake return None")
        code.I(LOAD_CONST, None)  # fake nxt returning None

    # now jump to the previously saved position
    code.c(f"goto saved pos")
    code.i(JUMP_ABSOLUTE, 0, jump_to=code.by_pos(p.pos + 2))

    code.c(f"---------------------")
    code.c(f"The original bytecode")
    code.c(f"---------------------")
    result = CodeType(
        0,
        0,
        0,
        len(code.co_varnames),
        new_stacksize + 1,
        flags,
        code.get_bytecode(),
        tuple(code.co_consts),
        tuple(code.co_names),
        tuple(code.co_varnames),
        f_code.co_filename,  # TODO: something smarter should be here
        f_code.co_name,
        f_code.co_firstlineno,  # TODO: this has to be fixed
        f_code.co_lnotab,
        )
    logging.info(f"resulting morph:\n{str(code)}")
    return result


def morph_stack(frame_data, root=True, **kwargs):
    """
    Morphs the stack.

    Parameters
    ----------
    frame_data : list
        States of all individual frames.
    root : bool
        Indicates if the stack contains a root
        frame where globals instead of locals
        need to be unpacked.
    kwargs
        Arguments to morph_execpoint.

    Returns
    -------
    result : CodeType, module
        The resulting morph for the root frame and
        the scope it belongs to.
    """
    prev = None
    for i, frame in enumerate(frame_data):
        logging.info(f"Preparing morph #{i:d}")
        prev = morph_execpoint(frame, prev,
            module_globals=frame_data if root and frame is frame_data[-1] else None,
            **kwargs), frame.scope
    return prev




"""
Frame object:

https://github.com/python/cpython/blob/3.9/Include/cpython/frameobject.h

Stack machine:

https://github.com/python/cpython/blob/3.9/Python/ceval.c
"""
from .py3_8 import *


def put_NULL(code):
    """
    BEGIN_FINALLY is gone
    """
    raise NotImplementedError("NULLs on stack are not implemented")




class NULLMeta(type):
    def __str__(cls):
        return "<NULL>"

    def __repr__(cls):
        return "NULL"


class NULL(metaclass=NULLMeta):
    """Represents NULL"""
    def __new__(cls):
        return cls




from dataclasses import dataclass, field
from types import CodeType, FunctionType
from itertools import count

import dis
locals().update(dis.opmap)


def long2bytes(l):
    result = tuple(map(int, l.to_bytes((l.bit_length() + 7) // 8, byteorder="big")))
    assert len(result) < 5
    if len(result) == 0:
        return 0,
    return result


def _trunc(s, l):
    if len(s) <= l:
        return s
    else:
        return s[:l-3] + "..."


@dataclass
class Instruction:
    """Represents a single opcode"""
    opcode: int
    arg: int
    pos: int = 0
    len: int = 2
    jump_to: object = None

    @property
    def is_jrel(self):
        return self.opcode in dis.hasjrel

    @property
    def is_jump(self):
        return self.opcode in dis.hasjabs

    @property
    def is_any_jump(self):
        return self.is_jrel or self.is_jump

    @property
    def pos_last(self):
        return self.pos + self.len - 2

    @property
    def bytes(self):
        arg_bytes = long2bytes(max(self.arg, 0))
        result = []
        for i in arg_bytes[:-1]:
            result.extend((EXTENDED_ARG, i))
        result.extend((self.opcode, arg_bytes[-1]))
        return bytes(result)

    def compute_jump(self, jx):
        if self.is_jrel:
            return self.arg * jx + self.pos_last + 2
        elif self.is_jump:
            return self.arg * jx
        else:
            return None

    def assert_valid(self, prev, lookup=None, jx=None):
        assert self.arg >= 0, f"arg is negative: {self.arg}"
        assert len(self.bytes) == self.len, f"len is invalid: len({repr(self.bytes)}) != {self.len}"
        if prev is not None:
            assert prev.pos + prev.len == self.pos, f"pos is invalid: {prev.pos}(prev.pos) + {prev.len}(prev.len) != {self.pos}(self.pos)"
        else:
            assert self.pos == 0, f"pos is non-zero: {self.pos}"
        if lookup is not None and self.is_any_jump:
            jump_points_to = lookup.get(self.compute_jump(jx), None)
            assert jump_points_to is self.jump_to, f"jump_to is invalid: {repr(self.jump_to)} vs {repr(jump_points_to)}"

    def __str__(self):
        return f"{self.pos:>6d} {_trunc(dis.opname[self.opcode], 18):<18} {self.arg:<16d}"

    def __repr__(self):
        return f"{dis.opname[self.opcode]}({self.arg}, pos={self.pos}, len={self.len})"

@dataclass
class Comment:
    """Represents a comment"""
    text: str
    @property
    def printable_text(self):
        if len(self.text) < 36:
            return self.text
        return f"{self.text[:32]}..."

    def __repr__(self):
        return f"       {self.printable_text}" + " " * (35 - len(self.text))


class CList(list):
    def index_store(self, x):
        try:
            return self.index(x)
        except ValueError:
            self.append(x)
            return len(self) - 1
    __call__ = index_store


class Bytecode(list):
    def __init__(self, opcodes, co_names, co_varnames, co_consts, jx=1):
        super().__init__(opcodes)
        self.pos = len(self)
        self.co_names = CList(co_names)
        self.co_varnames = CList(co_varnames)
        self.co_consts = CList(co_consts)
        self._jx = jx

    @classmethod
    def disassemble(cls, arg, **kwargs):
        if isinstance(arg, FunctionType):
            arg = arg.__code__
        code = arg.co_code
        marks = dis.findlinestarts(arg)
        try:
            lines = open(arg.co_filename, 'r').readlines()[arg.co_firstlineno - 1:]
            marks = list((i_opcode, i_line, lines[i_line - 1].strip()) for (i_opcode, i_line) in marks)
        except (TypeError, OSError, IndexError):
            marks = None
        result = cls([], arg.co_names, arg.co_varnames, arg.co_consts, **kwargs)
        arg = 0
        _len = 0
        for pos, (opcode, _arg) in enumerate(zip(code[::2], code[1::2])):
            arg = arg * 0x100 + _arg
            _len += 2
            if opcode != EXTENDED_ARG:
                start_pos = pos * 2 - _len + 2

                if marks is not None:
                    for i_mark, (mark_opcode, mark_lineno, mark_text) in enumerate(marks):
                        if mark_opcode <= start_pos:
                            result.c(f"{mark_lineno:<3d} {mark_text}")
                        else:
                            marks = marks[i_mark:]
                            break
                    else:
                        marks = []

                result.i(opcode, arg, start_pos, _len)
                arg = _len = 0
        result.eval_jumps()
        return result

    def i(self, opcode, arg=None, *args, **kwargs):
        if isinstance(opcode, Instruction):
            i = opcode
        else:
            i = Instruction(opcode, arg, *args, **kwargs)
        self.insert(self.pos, i)
        self.pos += 1
        return i

    def c(self, text):
        i = Comment(text)
        self.insert(self.pos, i)
        self.pos += 1
        return i

    def I(self, opcode, arg, *args, **kwargs):
        if opcode in dis.hasconst:
            return self.i(opcode, self.co_consts(arg), *args, **kwargs)
        elif opcode in dis.hasname:
            return self.i(opcode, self.co_names(arg), *args, **kwargs)
        elif opcode in dis.haslocal:
            return self.i(opcode, self.co_varnames(arg), *args, **kwargs)
        elif opcode in dis.hasjrel + dis.hasjabs:
            result = self.i(opcode, None)
            result.jump_to = arg
            return result
        else:
            raise ValueError(f"Unknown opcode: {dis.opname[opcode]}")

    def nop(self, arg):
        arg = bytes(arg)
        for i in arg:
            self.i(NOP, int(i))

    def iter_opcodes(self):
        for i in self:
            if isinstance(i, Instruction):
                yield i

    def by_pos(self, pos):
        for i in self.iter_opcodes():
            if i.pos == pos:
                return i
        else:
            raise ValueError(f"Instruction with pos={pos} not found")

    def eval_jumps(self):
        lookup = {i.pos: i for i in self.iter_opcodes()}
        for i in self.iter_opcodes():
            if i.is_any_jump and i.arg is not None:
                i.jump_to = lookup[i.compute_jump(self._jx)]

    def assign_pos(self):
        pos = 0
        for i in self.iter_opcodes():
            i.pos = pos
            pos += i.len

    def assign_jump_args(self):
        for i in self.iter_opcodes():
            if i.is_any_jump and i.jump_to is not None:
                if i.is_jump:
                    i.arg = i.jump_to.pos // self._jx
                elif i.is_jrel:
                    i.arg = (i.jump_to.pos - i.pos_last - 2) // self._jx

    def assign_len(self):
        for i in self.iter_opcodes():
            i.len = len(i.bytes)

    def assert_valid(self):
        prev = None
        lookup = {i.pos: i for i in self.iter_opcodes()}
        for opcode in self.iter_opcodes():
            opcode.assert_valid(prev, lookup=lookup, jx=self._jx)
            prev = opcode

    def get_bytecode(self):
        for i in range(5):
            self.assign_jump_args()
            self.assign_len()
            self.assign_pos()
            try:
                self.assert_valid()
                break
            except AssertionError:
                pass
        else:
            self.assert_valid()  # re-raise
        return b''.join(i.bytes for i in self.iter_opcodes())

    def __str__(self):
        lookup = {id(i): i_i for i_i, i in enumerate(self)}
        connections = []
        for i in self:
            connections.append([])
        for i_i, i in enumerate(self):
            if isinstance(i, Instruction) and i.jump_to is not None:
                i_j = lookup[id(i.jump_to)]
                i_mn = min(i_i, i_j)
                i_mx = max(i_i, i_j)
                occupied = set(sum(connections[i_mn:i_mx+1], []))
                for slot in count():
                    if slot not in occupied:
                        break
                for _ in range(i_mn, i_mx+1):
                    connections[_].append(slot)
        lines = []
        for i, c, c_prev, c_next in zip(self, connections, [[]] + connections[:-1], connections[1:] + [[]]):
            if len(c) == 0:
                lines.append(str(i))
            else:
                _str = []
                for _ in range(max(c) + 1):
                    if _ in c:
                        if _ in c_prev and _ in c_next:
                            _str.append("┃")
                        elif _ in c_prev and _ not in c_next:
                            _str.append("┛")
                        elif _ not in c_prev and _ in c_next:
                            _str.append("┓")
                        else:
                            _str.append("@")
                    else:
                        _str.append(" ")
                lines.append(str(i) + ''.join(_str))

            if isinstance(i, Instruction):
                represented, arg = _repr_arg(i.opcode, i.arg, self)
                if represented:
                    arg_repr = repr(arg)
                    if len(arg_repr) > 24:
                        arg_repr = f"<{type(arg).__name__} instance>"
                    lines[-1] = lines[-1] + f" ({arg_repr})"
        return '\n'.join(lines)


def _repr_arg(opcode, arg, code):
    if opcode in dis.hasconst:
        return True, code.co_consts[arg]
    elif opcode in dis.haslocal:
        return True, code.co_varnames[arg]
    elif opcode in dis.hasname:
        return True, code.co_names[arg]
    else:
        return False, arg


def _repr_opcode(opcode, arg, code):
    head = f"{dis.opname[opcode]:>20} {arg: 3d}"
    represented, val = _repr_arg(opcode, arg, code)
    if represented:
        return f"{head} {'(' + repr(val) + ')':<12}"
    else:
        return f"{head}" + " " * 13


def _dis(code_obj, alt=None):
    code = code_obj.co_code
    if alt is None:
        alt = code
    result = list(zip(code[::2], code[1::2], alt[::2], alt[1::2]))
    result_repr = []
    for i, (opc_old, arg_old, opc_new, arg_new) in enumerate(result):
        i *= 2
        if (opc_new, arg_new) == (opc_old, arg_old):
            result_repr.append((f"{i: 3d} {_repr_opcode(opc_new, arg_new, code_obj)}",))
        else:
            result_repr.append(("\033[94m", f"{i: 3d} {_repr_opcode(opc_new, arg_new, code_obj)} {_repr_opcode(opc_old, arg_old, code_obj)}", "\033[0m"))
    return result_repr


def cdis(code_obj, alt=None):
    return "\n".join(''.join(i) for i in _dis(code_obj, alt=alt))






import sys


module = __import__(f"py{sys.version_info[0]}_{sys.version_info[1]}", globals(), locals(), [], 1)
globals().update({
    i: getattr(module, i)
    for i in dir(module)
    if not i.startswith("_")
})

